{%raw%}
@charset "UTF-8";
@import url("https://fonts.googleapis.com/css?family=Open+Sans:300,400,600,700");

body {
    font-family: "Open Sans", sans-serif;
}

.clickable:hover {
    text-decoration: underline;
    color: blue;
}


.modeoption {
    margin: 5px;
}

.pending {
    color: #aaa;
    animation: color-transition 1s ease-in-out infinite alternate;
}
  
@keyframes color-transition {
    from {
        color: #aaa;
    }
    to {
        color: #eee;
    }
}

.actioning {
    animation: shadow-transition 1s ease-in-out infinite alternate;
}

@keyframes shadow-transition {
    from {
        text-shadow: 0px 0px 10px #aaaaaa15;
    }
    to {
        text-shadow: 0px 0px 10px #eee;
    }
}

.infobox-background {
    position: fixed; /* Stay in place */
    z-index: 1; /* Sit on top */
    padding-top: 100px; /* Location of the box */
    left: 0;
    top: 0;
    width: 100%; /* Full width */
    height: 100%; /* Full height */
    overflow: auto; /* Enable scroll if needed */
    background-color: rgb(0,0,0); /* Fallback color */
    background-color: rgba(0,0,0,0.4); /* Black w/ opacity */
}

.infobox-content {
    background-color: #fefefe;
    margin: auto;
    padding: 20px;
    border: 1px solid #888;
    width: 80%;
    top: 10%;
    z-index: 2;
    position: fixed;
    overflow-y: scroll;
    max-height: 80%;
}

.infobox-close {
    color: #aaaaaa;
    float: right;
    font-size: 28px;
    font-weight: bold;
}

.infobox-close:hover, .infobox-close:focus {
    color: #000;
    text-decoration: none;
    cursor: pointer;
}

.infobox-device-refresh {
    float: right;
    font-size: 28px;
}

#offline {
    height: 100vh;
    display: flex;
    justify-content: center;
    font-size: 48px;
    color: white;
}

#refresh-buttons {
    position: fixed;
    right: 10px;
    z-index: 1;
}

#refresh-buttons p {
    margin-top: 0;
}

#console {
    border: 1px solid black;
    background-color:rgba(179, 179, 179, 0.54);
    bottom: 0;
    left: 0;
    width: 40%;
    height: 1.5em;
    position: fixed;
}

#console.expanded {
    height: 40%;
}

#console .top {
    background-color: rgba(43, 43, 43, 0.397);
    top: 0;
    width: 100%;
    position: absolute;
    height: 1.5em;
}

#console .top .expanded-toggle {
    display: block;
    right: 1em;
    top: -.7em;
    position:absolute;
}

#request-list {
    margin-top: 2em;
}

#request-list .entry {
    border-bottom: 1px black dashed;
}

#request-list .entry em {
    font-size: 6pt;
    overflow-wrap: anywhere;
    display: block;
    opacity: 0.8;
}

#request-list .entry .elapsedTime {
    float: right;
    position: relative;
    top: -2.2em;
    right: 1em;
    font-size: 10pt;
    color: rgba(49, 49, 102, 0.795);
}

#request-list .entry .outgoing, #request-list .entry .incoming,  #request-list .entry .action-result {
    font-family: 'Andale Mono', 'Courier New', Courier, monospace;
    font-size: 9pt;
    overflow-wrap: anywhere;
    display: block;
}

#request-list .entry .outgoing::before {
    content: "➡";
}

#request-list .entry .incoming::before {
    content: "⬅";
}

#request-list .entry .action-result::before {
    content: "✪";
}

.sched-table tr td, .sched-table tr th {
    border: 1px solid #aaa;
}

td input.ng-invalid.ng-dirty {
    background-color: rgb(255, 152, 152);
}

.component-box {
    border: 1.5px solid #b3c6e0;
    border-radius: 12px;
    padding: 18px 20px 14px 20px;
    background: #f8fbff;
    margin-bottom: 24px;
    box-shadow: 0 2px 8px rgba(60, 80, 120, 0.06);
    /* max-width: 420px; */ /* Remove this line */
    width: 100%;             /* Add this line */
    box-sizing: border-box;  /* Ensure padding/border included in width */
}

.component-title {
    background: #3a6ea5;
    color: #fff;
    border-radius: 8px 8px 0 0;
    margin: -18px -20px 12px -20px;
    padding: 10px 18px;
    font-size: 1.15em;
    font-weight: 600;
    letter-spacing: 0.5px;
}

.component-doc {
    font-size: 0.96em;
    font-style: italic;
    color: #4a5a6a;
    margin-bottom: 8px;
}

.component-doc p {
    margin: 0 0 4px 0;
    line-height: 1.3;
}

.last-signal-value {
    margin-top: 10px;
    font-weight: 500;
    color: #2d4a6a;
}

table.arglist {
    display: table;
    width: 100%;
    border-collapse: collapse;
}
tr.argrow {
    display: table-row;
}
td.arg-name-cell {
    display: table-cell;
    vertical-align: top;
    padding: 8px 10px;
    width: 150px;
}

.arg-name {
    font-weight: 600;
}

.arg-description {
    font-size: 0.6em;
    font-style: italic;
    color: #817d83ff;
    margin-top: 2px;
}

tr.argrow:not(:last-child) {
    border-bottom: 1px solid #d6dbe8;
}

/* Publish button styling */
.publish-button {
    appearance: none;
    background: linear-gradient(135deg, #3a6ea5 0%, #427bb8 100%);
    color: #fff;
    border: 1px solid #2f5c89;
    border-radius: 6px;
    padding: 6px 14px 7px 14px;
    font-size: 0.85em;
    font-weight: 600;
    letter-spacing: 0.4px;
    cursor: pointer;
    box-shadow: 0 2px 4px rgba(38, 62, 92, 0.18), 0 1px 0 rgba(255,255,255,0.25) inset;
    text-shadow: 0 1px 1px rgba(0,0,0,0.25);
    transition: background .18s ease, box-shadow .18s ease, transform .08s ease;
    position: relative;
}
.publish-button:hover:not(:disabled) {
    background: linear-gradient(135deg, #427bb8 0%, #4c89c9 100%);
    box-shadow: 0 3px 7px rgba(38, 62, 92, 0.28), 0 1px 0 rgba(255,255,255,0.3) inset;
}
.publish-button:active:not(:disabled) {
    background: #2f5c89;
    transform: translateY(1px) scale(0.985);
    box-shadow: 0 2px 4px rgba(38, 62, 92, 0.25) inset;
}
.publish-button:focus-visible {
    outline: none;
    box-shadow: 0 0 0 3px rgba(255,255,255,0.85), 0 0 0 5px rgba(58,110,165,0.55), 0 2px 4px rgba(38, 62, 92, 0.25);
}
.publish-button:disabled {
    background: #b5c8da;
    border-color: #9ab1c7;
    color: #eef3f8;
    cursor: not-allowed;
    box-shadow: none;
    text-shadow: none;
    opacity: 0.85;
}
/* Subtle loading / pending state */
.publish-button.pending {
    position: relative;
    color: #eef6ff;
    animation: pulse-bright 1.1s ease-in-out infinite;
}
@keyframes pulse-bright {
    0% { box-shadow: 0 0 0 0 rgba(66,123,184,0.55); }
    70% { box-shadow: 0 0 0 6px rgba(66,123,184,0); }
    100% { box-shadow: 0 0 0 0 rgba(66,123,184,0); }
}
/* Optional subtle variant for destructive actions (not requested but available) */
.publish-button.danger {
    background: linear-gradient(135deg, #b83f3f 0%, #c64d4d 100%);
    border-color: #933131;
}
.publish-button.danger:hover:not(:disabled) {
    background: linear-gradient(135deg, #c64d4d 0%, #d85b5b 100%);
}
.publish-button.danger:active:not(:disabled) {
    background: #933131;
}


{%endraw%}